/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.enhancer.interceptors;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.easybeans.api.EasyBeansInvocationContext;
import org.ow2.easybeans.api.bean.EasyBeansBean;
import org.ow2.easybeans.api.interceptor.EZBInterceptorInvoker;
import org.ow2.easybeans.api.interceptor.EZBInterceptorManager;
import org.ow2.easybeans.api.interceptor.EZBInterceptorManagerFactory;
import org.ow2.easybeans.api.interceptor.EZBInvocationContextFactory;
import org.ow2.easybeans.deployment.metadata.ejbjar.EasyBeansEjbJarClassMetadata;
import org.ow2.easybeans.deployment.metadata.ejbjar.EasyBeansEjbJarMethodMetadata;
import org.ow2.easybeans.enhancer.lib.MethodRenamer;
import org.ow2.util.ee.metadata.ejbjar.api.IJClassInterceptor;
import org.ow2.util.ee.metadata.ejbjar.api.InterceptorType;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.metadata.structures.JMethod;

/**
 * This class will manage the creation of the invocation context objects.
 * @author Florent Benoit
 */
public class EasyBeansInvocationContextFactory implements EZBInvocationContextFactory {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(EasyBeansInvocationContextFactory.class);

    /**
     * Map between the signature of a method and all invokers defined for this
     * method.
     */
    private Map<String, List<EZBInterceptorInvoker>> methodsInterceptorInvokerList;

    /**
     * Map between the signature of a method and its java.lang.reflect.Method
     * object. (cache)
     */
    private Map<String, Method> methods;

    /**
     * List of classes that needs to be instantiate in the interceptor manager.
     */
    private List<String> interceptorClasses = null;

    /**
     * ClassLoader used to load classes.
     */
    private ClassLoader classLoader = null;

    /**
     * Build a new InvocationContext factory by analyzing each business method
     * and lifecycle method and creating the invokers.
     * @param classMetadata the metadata of the bean class that will be analyzed
     * @param classLoader the
     */
    public EasyBeansInvocationContextFactory(final EasyBeansEjbJarClassMetadata classMetadata, final ClassLoader classLoader) {
        this.classLoader = classLoader;

        // Init Map/List
        this.methodsInterceptorInvokerList = new HashMap<String, List<EZBInterceptorInvoker>>();
        this.methods = new HashMap<String, Method>();
        this.interceptorClasses = new ArrayList<String>();

        // Analyze each business method and build the invokers for each method /
        // bean method.
        Collection<EasyBeansEjbJarMethodMetadata> methodsMetadata = classMetadata.getMethodMetadataCollection();

        // Compute interceptors for each business method
        for (EasyBeansEjbJarMethodMetadata methodMetadata : methodsMetadata) {
            if (methodMetadata.isBusinessMethod()) {
                // Get list of interceptors
                List<IJClassInterceptor> allInterceptors = new MethodInterceptorsBuilder(methodMetadata,
                        InterceptorType.AROUND_INVOKE).getAllInterceptors();

                // build invokers
                List<EZBInterceptorInvoker> invokers = new ArrayList<EZBInterceptorInvoker>();

                String methodSignature = MethodHelper.getSignature(methodMetadata);

                // Store the invoker
                this.methodsInterceptorInvokerList.put(methodSignature, invokers);

                for (IJClassInterceptor interceptor : allInterceptors) {
                    String classname = interceptor.getClassName();
                    JMethod jMethod = interceptor.getJMethod();

                    // interceptor on the bean or outside ?
                    if (classMetadata.getClassName().equals(classname)) {
                        // interceptor is in the bean
                        // Add the invoker
                        invokers.add(new BeanInterceptorInvokerImpl(classname, jMethod, classLoader));
                    } else {
                        // outside of the bean
                        // Add the invoker
                        invokers.add(new StandaloneInterceptorInvokerImpl(classname, jMethod, classLoader));

                        // An interceptor instance will be required for this
                        // interceptor class
                        addInterceptorClass(classname);
                    }
                }

                // Add the method invocation call
                JMethod interceptedMethod = methodMetadata.getJMethod();
                JMethod originalMethod = new JMethod(interceptedMethod.getAccess(), MethodRenamer.encode(interceptedMethod
                        .getName()), interceptedMethod.getDescriptor(), interceptedMethod.getSignature(), interceptedMethod
                        .getExceptions());
                invokers.add(new BeanBusinessMethodInvokerImpl(classMetadata.getClassName(), originalMethod, classLoader));

                // Put in cache the business methods of the bean
                this.methods.put(methodSignature, MethodHelper.getMethod(classMetadata.getClassName(), methodMetadata
                        .getJMethod(), classLoader));
            }
        }

    }

    /**
     * Add the given classname to the list of the classes that will be managed
     * by the interceptor manager.
     * @param classname the name of the class
     */
    protected void addInterceptorClass(final String classname) {
        // Add it if class is not yet in the list
        if (!this.interceptorClasses.contains(classname)) {
            this.interceptorClasses.add(classname);
        }
    }

    /**
     * Gets the interceptor manager factory.
     * @return an instance of the manager factory
     */
    public EZBInterceptorManagerFactory getInterceptorManagerFactory() {
        return new EasyBeansInterceptorManagerFactory(this.interceptorClasses, this.classLoader);
    }

    /**
     * Gets an invocation context for the given method.
     * @param instance the bean's instance
     * @param interceptorManager the manager of the interceptors
     * @param methodSignature the key in order to find data for the given method
     * @param parameters the parameters of the method
     * @return an invocation context implementation
     */
    public EasyBeansInvocationContext getContext(final EasyBeansBean instance, final EZBInterceptorManager interceptorManager,
            final String methodSignature, final Object... parameters) {
        LOGGER.debug("Calling getContext for instance ''{0}'', interceptor manager ''{1}'', signature ''{2}'' and parameters ''{3}'' with interceptor invoker list set to ''{4}''", instance, interceptorManager, methodSignature, parameters, this.methodsInterceptorInvokerList);
        return new DynamicInvocationContextImpl(instance, this.methodsInterceptorInvokerList.get(methodSignature),
                interceptorManager, this.methods.get(methodSignature), parameters);
    }

}
